/*----------------------------------------------------------------------------
 *      Name:    sysCfg.c
 *      Purpose: SystemConfiguration
 *
 *	SystemCfg  the system configuration is stored in EEPROM. User parameters
 *	and system parameters are assigned static physical addresses, and a Checksum is
 *	performed on the memory area. The first byte is the system configuration version
 *	number. The last 2 bytes of EEPROM are the diagnostic test area (DTA), and the
 *	2 byte preceding the DTA  are the Checksum value. Help information may also be
 *	stored in a unique area of eeprom, with its own Checksum.
 *
 *	EEPROM Image
 *
 *      Copyright (c) 2008 SoftwareFlair Inc.
 *---------------------------------------------------------------------------*/
#include "type.h"
#include <string.h>
#include "eeprom.h"
#include "config.h"
#include "timers.h"

ushort chkSumCalc(ushort myStartAddr, ushort myEndAddr);
extern char verifySysCfg(EE_AREA_TYPE);
extern void getSysCfg(EE_AREA_TYPE);
extern void saveCfgDefaults(EE_AREA_TYPE);
extern void setShadowRegsToDef(EE_AREA_TYPE);

VOLATILE_REGISTERS shadowRegs;

SYS_CFG_SHADOW_REG sysCfgShadowRegs;	// these shadow registers are modified by the code, when a change is required
SYS_CFG_SHADOW_REG sysCfgEeprom;			// this structure contains what is in eeprom
CAL_SHADOW_REG calShadowRegs;	// these shadow registers are modified by the code, when a change is required
CAL_SHADOW_REG calEeprom;			// this structure contains what is in eeprom
static ulong myStartAddr[EE_AREA_END];
static ulong myChkSmAddr[EE_AREA_END];
static ulong myShadowRegSize[EE_AREA_END] = {0,0};
static uchar* myShadowPtr[EE_AREA_END] = {0,0};
static uchar* myEePtr[EE_AREA_END] = {0,0};

short isCalImageGood = FALSE;
short isSysCfgImageGood = FALSE;
ulong currentSysCfgBank = EE_SYSCFG0_START; 	// current System Configuration Bank: 0 or 1 - double buffered

void sysCfgInit(void)
{
	// initailize System Configuration Image Areas
	myStartAddr[EE_AREA_SYS_CFG] = (ulong) currentSysCfgBank;
	myChkSmAddr[EE_AREA_SYS_CFG] = currentSysCfgBank + EE_SYSCFG0_CHKSUM;
	myShadowRegSize[EE_AREA_SYS_CFG] = sizeof(sysCfgShadowRegs);
	myShadowPtr[EE_AREA_SYS_CFG] = (uchar*) &sysCfgShadowRegs;
	myEePtr[EE_AREA_SYS_CFG] = (uchar*) &sysCfgEeprom;

	myStartAddr[EE_AREA_CAL] = (ulong) EE_CAL_START;
	myChkSmAddr[EE_AREA_CAL] = EE_CAL_CHKSUM;
	myShadowRegSize[EE_AREA_CAL] = sizeof(calShadowRegs);
	myShadowPtr[EE_AREA_CAL] = (uchar*) &calShadowRegs;
	myEePtr[EE_AREA_CAL] = (uchar*) &calEeprom;

	// is eeprom alive and well?
	if(isEepromWorking == TRUE)
	{
			////////////////
			// SYSTEM CONFIG
			if(verifySysCfg(EE_AREA_SYS_CFG) == TRUE)	// is this system config image/bank valid?
			{
				getSysCfg(EE_AREA_SYS_CFG);	// yes, get the image
				isSysCfgImageGood = TRUE;
			}
			else
			{
				// check the "other" System Configuration bank!
				if(currentSysCfgBank == EE_SYSCFG0_START)
					currentSysCfgBank = EE_SYSCFG1_START;
				else
					currentSysCfgBank = EE_SYSCFG0_START;

				myStartAddr[EE_AREA_SYS_CFG] = (ulong) currentSysCfgBank;
				myChkSmAddr[EE_AREA_SYS_CFG] = currentSysCfgBank + EE_SYSCFG0_CHKSUM;

				if(verifySysCfg(EE_AREA_SYS_CFG) == TRUE)	// is this 2nd system config image/bank valid?
				{
					getSysCfg(EE_AREA_SYS_CFG);	// yes, get the image
					isSysCfgImageGood = TRUE;
				}
				else	// both banks are bad!!
				{
					// FRG: Remove Later!!
					//putString(COMM_0, "*Init Sys Cfg!!\n");

					saveCfgDefaults(EE_AREA_SYS_CFG);	// no, save the system config

					if(verifySysCfg(EE_AREA_SYS_CFG) == TRUE)	// is the system config image valid?
					{
						getSysCfg(EE_AREA_SYS_CFG);	// yes, get the image
						isSysCfgImageGood = TRUE;
					}
					else
					{
						setShadowRegsToDef(EE_AREA_SYS_CFG);
						//errLogFlagSet(ERR_EEPROM_ERROR);

						// FRG: Remove Later!!
						//putString(COMM_0, "*Sys Cfg Failed!!\n");
					}
				}
			}

			/////////////
			// CAL
			if(verifySysCfg(EE_AREA_CAL) == TRUE)	// is the calibration image valid?
			{
				getSysCfg(EE_AREA_CAL);	// yes, get the image
				isCalImageGood = TRUE;
			}
			else
			{
				// FRG: Remove Later!!
				//putString(COMM_0, "*Cal Failed!!\n");
				isCalImageGood = FALSE;
			}			
	}
	else		// eeprom issues!
	{
		setShadowRegsToDef(EE_AREA_SYS_CFG);
		setShadowRegsToDef(EE_AREA_CAL);

		//errLogFlagSet(ERR_EEPROM_ERROR);

		//putString(COMM_0, "\nEeprom Failed!!!\n");
		//putString(COMM_1, "\nEeprom Failed!!!\n");

 	}


 	if(isSysCfgImageGood == TRUE && isCalImageGood == TRUE)
 	{
		//If the EEPROM DEF version in this unit was generated by a previous firmware build
		if(sysCfgShadowRegs.version < EE_DEF_VERSION)
		{
//			if(sysCfgShadowRegs.version < 0x01)
//			{
//				
//			}
//And so on...
//			if(sysCfgShadowRegs.version < 0x02)
//			{
//		
//			}

			saveSysCfgDelta(EE_AREA_CAL);

			sysCfgShadowRegs.version = EE_DEF_VERSION; //Update the EEPROM DEFINITION version to make current.
			saveSysCfgDelta(EE_AREA_SYS_CFG);
		}
	}
}

// force the system configuration values to default in the eeprom
void saveCfgDefaults(EE_AREA_TYPE myArea)
{
	ushort myCnt;
	uchar* myPtr = 0;
	ushort myCheckSum;
	uchar myChkSumByte; 

	setShadowRegsToDef(myArea);

	myPtr = (uchar*) myShadowPtr[myArea];

	for(myCnt = myStartAddr[myArea]; myCnt < (myStartAddr[myArea] + myShadowRegSize[myArea]); myCnt++)
	{
		// write all of the system configuration values from shadow registers into EEPROM
		eepromWrite(myCnt, myPtr++,1);
	}

	// calculate checksum
	myCheckSum = chkSumCalc(myStartAddr[myArea], myChkSmAddr[myArea] - 1);

	myChkSumByte=myCheckSum >> 8;

	// save checksum
	eepromWrite(myChkSmAddr[myArea], &myChkSumByte,1); //MSB

	myChkSumByte=myCheckSum & 0x00FF;

	eepromWrite(myChkSmAddr[myArea] + 1, &myChkSumByte,1); //LSB
}

// save the system config shadow registers to eeprom
void saveSysCfg(EE_AREA_TYPE myArea)
{
	ushort myCnt;
	uchar* myPtr;
	uchar myData;
	ushort myCheckSum = 0;
	ulong tmpCurrentSysCfgBank;
	uchar myChkSumByte; 

	// double banking is used for the system config area
	// start to write the new transaction, when a full bank write is complete with checksum, then switch banks, and destroy the previous image
	if(myArea == EE_AREA_SYS_CFG)
	{
		// check the "other" System Configuration bank!
		if(currentSysCfgBank == EE_SYSCFG0_START)
			tmpCurrentSysCfgBank = EE_SYSCFG1_START;
		else
			tmpCurrentSysCfgBank = EE_SYSCFG0_START;

		myStartAddr[EE_AREA_SYS_CFG] = (ulong) tmpCurrentSysCfgBank;
		myChkSmAddr[EE_AREA_SYS_CFG] = (ulong) tmpCurrentSysCfgBank + EE_SYSCFG0_CHKSUM;
	}

	myPtr =  myShadowPtr[myArea];

	for(myCnt = myStartAddr[myArea]; myCnt < (myStartAddr[myArea] + myShadowRegSize[myArea]); myCnt++)
	{
		// write all of the system configuration values from shadow registers into EEPROM
		eepromWrite(myCnt, myPtr++,1);
	}

	// calculate checksum
	myCheckSum = chkSumCalc(myStartAddr[myArea], myChkSmAddr[myArea] - 1);

	myChkSumByte = myCheckSum >> 8;

	// save checksum
	eepromWrite(myChkSmAddr[myArea], &myChkSumByte,1); //MSB

	myChkSumByte = myCheckSum & 0x00FF;

	eepromWrite(myChkSmAddr[myArea] + 1, &myChkSumByte,1); //LSB

 	// double banking is used for the system config area
	if(myArea == EE_AREA_SYS_CFG)
	{
		// destroy the previous bank.
		eepromRead(tmpCurrentSysCfgBank + EE_SYSCFG0_CHKSUM + 1, &myData,1);	 // read the previous "correct" checksum LSB value
		// destroy the image area: save correct checksum + 1. This allows the image area to be analyzed in the future
		myData+=1;
		eepromWrite(tmpCurrentSysCfgBank + EE_SYSCFG0_CHKSUM + 1, &myData + 1,1);

		currentSysCfgBank = tmpCurrentSysCfgBank;	// switch to the new bank.
	}
}

// save the system config shadow register(s) that do not match what is in the eeprom
void saveSysCfgDelta(EE_AREA_TYPE myArea)
{
	ushort myCnt;
	uchar* myPtrShReg;
	uchar* myPtrEe;
	uchar myData;
	//short eePromChangeMade = FALSE;
	//ushort myLen;
	ushort myCheckSum = 0;
	ulong tmpCurrentSysCfgBank;
	uchar myChkSumByte; 

	// double banking is used for the system config area
	// start to write the new transaction, when a full bank write is complete with checksum, then switch banks, and destroy the previous image
	if(myArea == EE_AREA_SYS_CFG)
	{
		// check the "other" System Configuration bank!
		if(currentSysCfgBank == EE_SYSCFG0_START)
			tmpCurrentSysCfgBank = EE_SYSCFG1_START;
		else
			tmpCurrentSysCfgBank = EE_SYSCFG0_START;

		myStartAddr[EE_AREA_SYS_CFG] = (ulong) tmpCurrentSysCfgBank;
		myChkSmAddr[EE_AREA_SYS_CFG] = (ulong) tmpCurrentSysCfgBank + EE_SYSCFG0_CHKSUM;
	}

	myPtrShReg = myShadowPtr[myArea];
	myPtrEe = myEePtr[myArea];

	// read in the eeprom contents for the sys cfg new bank.
	for(myCnt = myStartAddr[myArea]; myCnt < (myStartAddr[myArea] + myShadowRegSize[myArea]); myCnt++)
	{
		// read all of the system configuration values from EEPROM into shadow registers
		eepromRead(myCnt, myPtrEe++,1);
		delayUs(20);

	}

	myPtrEe = myEePtr[myArea];

	for(myCnt = myStartAddr[myArea]; myCnt < ( myStartAddr[myArea] + myShadowRegSize[myArea]); myCnt++)
	{
		// if the shadow register has changed (does not match what is in eeprom), write the "new" correct value to the EEProm
		if(*myPtrShReg	!= *myPtrEe)
		{
			// yes, write the system configuration value from shadow registers into EEPROM
			eepromWrite(myCnt, myPtrShReg,1);
			*myPtrEe = *myPtrShReg;		// save the new value in the eeprom variable
		}

		// move onto the next byte/attribute for comparison
		myPtrShReg++;
		myPtrEe++;
	}

	// calculate checksum, DMS: Should be able to calculate checksum for myEePtr[myArea] since it contains all mods
	//instead of reading the entire image back in again. 
	myCheckSum = chkSumCalc(myStartAddr[myArea], myChkSmAddr[myArea] - 1);
		
	// save checksum
	myChkSumByte = myCheckSum >> 8;	
	eepromWrite(myChkSmAddr[myArea], &myChkSumByte,1);
	
	myChkSumByte = myCheckSum & 0x00FF;
	eepromWrite(myChkSmAddr[myArea] + 1, &myChkSumByte,1);

 	// double banking is used for the system config area
	if(myArea == EE_AREA_SYS_CFG)
	{
		// destroy the previous bank.
		eepromRead(currentSysCfgBank + EE_SYSCFG0_CHKSUM + 1, &myData,1);	 // read the previous "correct" checksum LSB value

		delayUs(20);

		myData+=1;

		// destroy the image area: save correct checksum + 1. This allows the image area to be analyzed in the future
		eepromWrite(currentSysCfgBank + EE_SYSCFG0_CHKSUM + 1, &myData,1);

		currentSysCfgBank = tmpCurrentSysCfgBank;	// switch to the new bank.
	}
}

// verify the saved system configuration image
char verifySysCfg(EE_AREA_TYPE myArea)
{
	ushort myCheckSum = 0;
	uchar myData;
	ushort myCalcCheckSum = 0;

	// calculate checksum
	myCalcCheckSum = chkSumCalc(myStartAddr[myArea], myChkSmAddr[myArea] - 1);

	// get checksum
	eepromRead(myChkSmAddr[myArea], &myData,1);
	delayUs(20);

	myCheckSum = myData << 8;
	eepromRead(myChkSmAddr[myArea] + 1, &myData,1);
	delayUs(20);

	myCheckSum |= myData;

	if(myCalcCheckSum != myCheckSum)
		return FALSE;

	return TRUE;
}

// get system configuration values from the eeprom.
void getSysCfg(EE_AREA_TYPE myArea)
{
	ushort myCnt;
	uchar* myPtr;
	//ushort myLen;

	myPtr = myEePtr[myArea];

	for(myCnt = myStartAddr[myArea]; myCnt < (myStartAddr[myArea] + myShadowRegSize[myArea]); myCnt++)
	{
		// read all of the system configuration values from EEPROM into shadow registers
		eepromRead(myCnt, myPtr++,1);
		delayUs(20);
	}
	memcpy(myShadowPtr[myArea], myEePtr[myArea], myShadowRegSize[myArea]);	// copy the eeprom image into the shadow registers
}

void setShadowRegsToDef(EE_AREA_TYPE myArea)
{
	if(myArea == EE_AREA_SYS_CFG)
	{
		sysCfgShadowRegs.version = EE_DEF_VERSION;
		sysCfgShadowRegs.Echo = EE_DEF_ECHO;
		sysCfgShadowRegs.Prompt=EE_DEF_PROMPT;
		
	}
	else if(myArea == EE_AREA_CAL)
	{
		calShadowRegs.calVersion = EE_CAL_VER;
		calShadowRegs.laserTypeA = EE_DEF_LSR_TYPES;
		calShadowRegs.laserTypeB = EE_DEF_LSR_TYPES;
		calShadowRegs.laserTypeC = EE_DEF_LSR_TYPES;
		calShadowRegs.laserTypeD = EE_DEF_LSR_TYPES;
		calShadowRegs.maxPwrLsr_A = EE_DEF_MAX_PWR;
		calShadowRegs.maxPwrLsr_B = EE_DEF_MAX_PWR;
		calShadowRegs.maxPwrLsr_C = EE_DEF_MAX_PWR;
		calShadowRegs.maxPwrLsr_D = EE_DEF_MAX_PWR;
		calShadowRegs.nomLsrPwr_A = EE_DEF_NOM_PWR;
		calShadowRegs.nomLsrPwr_B = EE_DEF_NOM_PWR;
		calShadowRegs.nomLsrPwr_C = EE_DEF_NOM_PWR;
		calShadowRegs.nomLsrPwr_D = EE_DEF_NOM_PWR;
		calShadowRegs.nomWaveLen_A = EE_DEF_NOM_WAVE;
		calShadowRegs.nomWaveLen_B = EE_DEF_NOM_WAVE;
		calShadowRegs.nomWaveLen_C = EE_DEF_NOM_WAVE;
		calShadowRegs.nomWaveLen_D = EE_DEF_NOM_WAVE;
		strcpy(calShadowRegs.lsrPartNum_A,EE_DEF_PART_NUM);
		strcpy(calShadowRegs.lsrPartNum_B,EE_DEF_PART_NUM);
		strcpy(calShadowRegs.lsrPartNum_C,EE_DEF_PART_NUM);
		strcpy(calShadowRegs.lsrPartNum_D,EE_DEF_PART_NUM);
		strcpy(calShadowRegs.lsrSerialNum_A,EE_DEF_SER_NUM);
		strcpy(calShadowRegs.lsrSerialNum_B,EE_DEF_SER_NUM);
		strcpy(calShadowRegs.lsrSerialNum_C,EE_DEF_SER_NUM);
		strcpy(calShadowRegs.lsrSerialNum_D,EE_DEF_SER_NUM);
		strcpy(calShadowRegs.lsrBeamShape_A,EE_DEF_BM_SHAPE);
		strcpy(calShadowRegs.lsrBeamShape_B,EE_DEF_BM_SHAPE);
		strcpy(calShadowRegs.lsrBeamShape_C,EE_DEF_BM_SHAPE);
		strcpy(calShadowRegs.lsrBeamShape_D,EE_DEF_BM_SHAPE);
	}
}

void forceSysCfgDefaults(void)
{
	saveCfgDefaults(EE_AREA_SYS_CFG);	// save the system config
	saveCfgDefaults(EE_AREA_CAL);			// save the calibration information
}

// The end address is part of the Checksum calculation!!
ushort chkSumCalc(ushort myStartAddr, ushort myEndAddr)
{
	ushort myCheckSum = 0;
	ushort mySize;
	uchar myData;

	mySize = (myEndAddr + 1) - myStartAddr;

	while(mySize--)
	{
		eepromRead((ushort) myStartAddr++, &myData,1);
		delayUs(20);
		myCheckSum += myData;
	}  

	return(myCheckSum);
}


